function decode_peer(data)
	return ( data:byte(1).."."..data:byte(2).."."..data:byte(3).."."..data:byte(4) ),
		   ( data:byte(5)*256 + data:byte(6) )
end

function ipv4_array(address)
	local ip = {}
	ip[1], ip[2], ip[3], ip[4] = string.match(address, "([0-9]+).([0-9]+).([0-9]+).([0-9]+)")
	if ip[1] then
		ip[1], ip[2], ip[3], ip[4] = tonumber(ip[1]), tonumber(ip[2]), tonumber(ip[3]), tonumber(ip[4])
		return ip
	end
end

function encode_ipv4(ip)
	local a1, a2, a3, a4 = string.match(ip, "([0-9]+).([0-9]+).([0-9]+).([0-9]+)")
	return string.char(a1)..string.char(a2)..string.char(a3)..string.char(a4)
end

function encode_port(port, le)
	local p1 = math.mod(port, 256)
	local p2 = (port - p1) / 256
	if le then p1, p2 = p2, p1 end
	return string.char(p2)..string.char(p1)
end


function encode_peer_le(address, port)
	return encode_ipv4(address)..encode_port(port, true)
end

function encode_peer(address, port, le)
	return encode_ipv4(address)..encode_port(port)
end

function decode_node(data, node_index)
	node_index = node_index or 1
	local s  = (26 * node_index) - 25
	if #data < (s + 25) then return nil end
	return data:sub(s, s + 19), decode_peer(data:sub(s + 20, s + 25))
end

function encode_node(node)
	return node.id..encode_peer(node.address, node.port)
end