function next_client(tcp_port)
	local client, err = tcp_port:accept()
	
	if not client then
		if (err ~= "timeout") then
			print("accept error: "..err)
			debug.debug()
		end
		return 
	end
	
	local line = client:receive("*l")
	if not line then client:close() return end
	
	if string.match(line, "GNUTELLA CONNECT/0.6") then
		return g2_main(client)
	end
	
	-- Brouser --
	if string.match(line, " / ") then
		return main_page(client)
	end
	
	if string.match(line, " /console ") then
		to_http(nil, "text/html")
		local function print_fnc(line)
			if not client:send(line.."\n") then
				client:close()
				remove_print_fnc(print_fnc)
			end
		end
		add_print_fnc(print_fnc)
		return
	end
	
	if string.match(line, " /info ") then
		return http_info(client)
	end
	
	if string.match(line, " /map ") then
		return svg_map(client)
	end	
	
	
	-- Tracker --
	local info_hash = string.match(line, "[?&]info_hash=([^& ]+)")
	if not info_hash then 
		client:close()
		return
	else
		info_hash = unescape(info_hash)
		if #info_hash ~= 20 then
			print("#info_hash", #info_hash, info_hash)
			client:close()
			return
		end
	end
	
	local port = string.match(line, "port=([0-9]+)")
	port = (port and tonumber(port))
	local responce
	local table_type
	
	local header = client:receive("*l")
	while header and #header > 0 do
		local a, p = string.match(header, "Listen%-IP: ([0-9%.]+):([0-9]+)")
		port = port or (p and tonumber(p))
		header = client:receive("*l")
	end
	
	if string.find(line, " /uri-res", 1, true) then
		print("\t\t\t> http (uri-res)")

		local link = string.match(line, " ([^ ]+) ")
		
		local buff = new_string_builder()
		buff.add([[
<html>
	<head>
		<title>search by urn</title>
	</head>
	<body>
<p>This page refresh every 120 seconds. Whait plz.</p>
<p>
]])


		
		local peers = get_peers(info_hash, port)
		local x_alt = ""
		if peers then
			local peers_x_alt = new_string_builder()
			for compact, peer in pairs(peers) do
				if not peer.x_alt_sended then
					local a, p = decode_peer(compact)
					peers_x_alt.add(a..":"..p)
					peer.x_alt_sended = os.time()
					buff.add([[<a href="http://]]..a..":"..p..link..[[">]]
								..a..":"..p.."</a><br>\n")
				end
			end
			if not peers_x_alt.empty() then
				x_alt = "X-Alt: "..peers_x_alt.get(",").."\n"
				
			end
		end


		buff.add([[
		</p>
	</body>
</html>
]])
		local content = buff.get()
		client:send([[
HTTP/1.1 503 Bisy
Server: ]]..script_name.."\n"..[[
X-Queue: position=2,length=10,limit=20,pollMin=10,pollMax=100");
Retry-After: 120
Refresh: 120]]
.."\n"..x_alt.."Content-Length: "..#content.."\n\n"..content)

		client:close()
		return
	
	elseif string.find(line, " /announce", 1, true) then
		print("\t\t\t> http (announce)")
		local peers = get_peers(info_hash, port)
		local peers_compact = new_string_builder()
		if peers then
			for compact, peer in pairs(peers) do
				--if not peer.compact_sended then
					peers_compact.add(compact)
					--peer.compact_sended = os.time()
				--end
			end
		end
		responce = {interval = 120}
		if not peers_compact.empty() then
			responce.peers = peers_compact.get()
		end
		table_type = {[responce] = announce_responce}
	elseif string.find(line, " /scrape", 1, true) then
		print("\t\t\t> http (scrape)")
		responce = {[info_hash] = {complete = 0, downloaded = 0, incomplete = 0}}
		table_type = {[responce] = scrape_responce, [responce[info_hash]] = scrape_details}
	end
	
	client:send(to_http(to_bencode(responce, table_type)))
	client:close()
end

announce_responce = { t = "d", o = {"interval", "peers"} }
scrape_responce = { t = "d" }
scrape_details = {t = "d", o = {"complete","downloaded","incomplete"}}

ferrmsg = function(text)
	return "d14:failure reason"..#text..":"..text.."e"
end

standart_header = [[
HTTP/1.1 200 OK
Server: ]]..script_name.."\n"..[[
Connection: close
]]

function to_http(body, content_type)
	body = body or ferrmsg("tracker internal error", content_type)
	local data = standart_header
	data = data.."Content-Type: "..(content_type or "text/plain").."\n"
	if body then data = data.."Content-Length: "..string.len(body).."\n" end
	data = data.."\n"..(body or '')
	return data
end


function main_page(client)
		local buff = new_string_builder()
		buff.add("<html><head><title>Lua DHT Tracker Info</title></head><body>")
		buff.add("<pre>"..welcome_msg.."</pre>")
		buff.add(
[[
<p>
	node id: ]]..b32enc(my_id).."<br />\n"..[[
	nodes count: ]]..(nodes.count or "unknown").."\n"..[[
</p>

<p>
	<a href="/console">Console Out</a></br>
	<a href="/info">Hosted torrents</a></br>
	<a href="/map">Nodes Map</a></br>
</p>

 	
<p>
input btih and click this link:<br />
	<a href="" id="magnet">magnet:?xt=urn:btih:</a><input type="text" id="btih" onChange="document.getElementById('magnet').href='magnet:?xt=urn:btih:'+this.value+'&amp;tr=http://127.0.0.1:]]..port_number..[[/announce'; document.getElementById('magnet2').href=document.getElementById('magnet').href" /><a href="" id="magnet2">&amp;tr=http://127.0.0.1:]]..port_number..[[/announce </a>
</p>
]]		)
		buff.add("</table></body></html>")
		client:send(to_http(buff.get("\n"), "text/html"))
		client:close()
end

local maximum_ugol = 255 * 256 * 256 * 256

function spiral(ugol)
	--ugol = maximum_ugol - ugol
	local m1 = (ugol/maximum_ugol * 200)
	local m2 = (ugol/maximum_ugol * 46)
	return (50+math.cos(m1)*(m2)), (50+math.sin(m1)*(m2))
end

function svg_map(client)
	print("\t\t\t>http map")
	local buff = { new_string_builder(), new_string_builder(), new_string_builder(), 
	new_string_builder() }
	buff[1].add([[<?xml version="1.0" encoding="UTF-8"?>
<svg version="1.1"
 baseProfile="full"
 xmlns="http://www.w3.org/2000/svg"
 xmlns:xlink="http://www.w3.org/1999/xlink"
 xmlns:ev="http://www.w3.org/2001/xml-events"
 width="800" height="600">
]])
	
	
	buff[1].add([[<g stroke="red" >]])
	buff[2].add([[<g fill="#000088" >]])
	buff[3].add([[<g stroke="#000088" stroke-whith="2px" fill="none">]])
	buff[4].add([[<g stroke="#880000">]])
	
	for ip, port_list in pairs(nodes.ap) do
		for port, node in pairs(port_list) do
			local opacity = (1 - (os.time() - (node.last_seen or node.added)) / (15*60) )
			
			if opacity>0 then
				opacity = (svg_opacity and ' opacity="'..opacity..'" ') or ''
				
				local x, y = spiral(statistic_distance(my_id, node.id))

				local fill = (node.confirmed and ' fill="green" ') or (node.sended and ' fill="lime"') or ( (not node.last_seen) and ' fill="red" ' ) or ""
				
				buff[2].add([[<circle cx="]]..x..[[%" cy="]]..y..[[%" r="2px"]]
									..fill..opacity..[[ />]])
				
				if node.conected_to then
					local stroke = (node.confirmed and ' stroke="green" ') or (node.last_seen and ' stroke="bule" ') or ""
					local x2, y2 = spiral(statistic_distance(my_id, node.conected_to))
					buff[1].add('<line x1="'..x..'%" y1="'..y..'%" x2="'..x2..'%" y2="'..y2..'%"'..stroke..opacity..' />')
				end
				
				
				if node.announced then
					buff[3].add([[<circle cx="]]..x..[[%" cy="]]..y..[[%" r="5px"]]
									..opacity..[[ stroke="red" />]])
				elseif node.peer then
					buff[3].add([[<circle cx="]]..x..[[%" cy="]]..y..[[%" r="5px"]]
									..opacity..[[ />]])
				end
			end
		end
	end
	local  x, y =  spiral(0)
	
	buff[3].add([[<circle cx="]]..x..[[%" cy="]]..y..[[%" r="5px" stroke="green" />]])
	
	find_close_id(my_id, 50, 
		function(node, buff)
			local  x2, y2 =  spiral(statistic_distance(my_id, node.id))
			
			buff.add('<line x1="'..x..'%" y1="'..y..'%" x2="'..x2..'%" y2="'..y2..'%" />')
			
			x = x2
			y = y2
			return true
		end
		, {buff[4]}
	)
	
	buff[1].add("</g>")
	buff[2].add("</g>")
	buff[3].add("</g>")
	buff[4].add("</g>")
	buff[4].add("</svg>")
	client:send(to_http(buff[1].get("\n")..buff[2].get("\n")..buff[3].get("\n")..buff[4].get("\n"), "image/svg+xml"))
	client:close()
end

function sort_by_count(f, s)
	if f.count > s.count then
		return true
	elseif f.count < s.count then
		return false
	end
	
	return less(f.hash_raw, s.hash_raw, my_id)
end

function http_info(client)
	print("\t\t\t>http info")
	local buff1 = {}
	for hash, peer_list in pairs(peers.btih) do
		table.insert(buff1, { hash = b32enc(hash), hash_raw = hash, count = count_elements(peer_list) })
	end
	table.sort(buff1, sort_by_count)
	local buff = new_string_builder()
	buff.add("<html><head><title>Lua DHT Tracker Info</title></head><body><table>")
	buff.add("<tr><th>BitTorrent Info Hash</th><th>Peers count</th></tr>")
	for index, value  in pairs(buff1) do
		buff.add("<tr><td><a href='magnet:?xt=urn:btih:"..value.hash.."&amp;tr=http://127.0.0.1:"..port_number
							.."/announce'>"..value.hash.."</a></td><td>"..value.count.."</td></tr>")
	end
	buff.add("</table></body></html>")
	client:send(to_http(buff.get("\n"), "text/html"))
	client:close()
	return
end