
function is_dictionary(data)
	return data:sub(1,1) == "d" and data:sub(-1) == "e"
end

local numbers={["0"]=0,["1"]=1,["2"]=2,["3"]=3,["4"]=4,["5"]=5,["6"]=6,["7"]=7,["8"]=8,["9"]=9}

function from_bencode(data)
	local table_info = {}
	
	local function  parse(index)
		if index > #data then return end 
		local t = data:sub(index, index)
		--status_print(t)
		if t == "i" or numbers[t] then -- number or string
			local start, endindex, number = string.find(data, "i?([0-9]+)[e:]", index)
			
			assert(start == index)
			
			number = tonumber(number)
			
			if not number then return end
			
			if t == "i" then -- is number?
				return number, endindex 
			end
			
			if ( number < 0 ) or (math.floor(number) ~= number) or ( endindex + number > #data ) then return end
			local value = data:sub(endindex + 1, endindex + number)
			return value, endindex + number
		
		elseif t == "l" then -- list
			local value, endindex = parse(index + 1)
			local list={}
			--
			while value do
				table.insert(list,value)
				value, endindex = parse(endindex + 1)
			end
			
			if not endindex then return end
			
			return list, endindex
		elseif t == "d" then -- dictionary
			local dict={}
			local dict_info = { t = t, o = {} }
			table_info[dict] = dict_info
			local key, endindex = parse(index + 1)
			local value
			while key do
				value, endindex = parse(endindex + 1)
				if value then
					table.insert(dict_info.o, key)
					dict[key]=value
				else
					break
				end
				if not endindex then return end
				key, endindex = parse(endindex + 1)
				if not endindex then return end
			end
			return dict, endindex
		elseif t == "e" then
			return nil, index
		end
	end
	return parse(1), table_info
end

function to_bencode(value, table_info, buff)
	--print(type(value))
	if not buff then
		local buff = {}
		to_bencode(value, table_info, buff)
		return table.concat(buff)
	elseif type(value) == "number" then
		table.insert(buff, "i"..value.."e")
	elseif type(value) == "string" then
		table.insert(buff, #value..":"..value)
	elseif type(value) == "table" then
		local dict = table_info[value]
		table.insert(buff, ((dict and "d") or "l"))
		
		if not dict then
			for _, v in ipairs(value) do
				to_bencode(v, table_info, buff)
			end
		elseif dict then
			if dict.o then
				for _, k in ipairs(dict.o) do
					if value[k] then
						to_bencode(k , table_info, buff)
						to_bencode(value[k] , table_info, buff)
					end
				end
			else
				for k, v in pairs(value) do
					to_bencode(k , table_info, buff)
					to_bencode(v , table_info, buff)
				end
			end
		end
		table.insert(buff, "e")
	end
end

function set_value(tbl, table_info, value, key)
	local info = table_info[tbl]
	if key and value then
		if not info then 
			info = {t = "d", o = {}}
			table_info[tbl] = info
		end
		tbl[key] = value
		table.insert( info.o, key )
	elseif value and not (info or key) then 
		table.insert( tbl, value )
	end
end